# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Any, Dict

import jax
import jax.numpy as jnp


def apply_rope(
    # (seq_len, num_heads, head_dim)
    inputs: jax.Array,
    # (3, seq_len) for M-RoPE, otherwise (seq_len,)
    positions: jax.Array,
    head_dim: int,
    rope_theta: float = 10000,
    rope_scaling: Dict[str, Any] = None,
    rope_input_ordering: str = "split",
    *,
    rope_proportion: float = 1.0,
) -> jax.Array:
    """
    Applies Rotary Positional Embedding using the sine and cosine strategy.

    This implementation assumes the input tensor has a shape that might include
    padding on the last dimension (head_dim).
    RoPE is applied only to the first `head_dim` features, and the result is
    padded back to the original dimension if necessary.
    If rope_input_ordering is "split", then the input pairs for rotation are taken one from the
    first and one from the second half of the head_dim. If it is "interleaved" then
    adjacent values are used as inputs for rotation.
    """

    # M-RoPE support for Qwen2.5-VL
    if positions.ndim == 2 and positions.shape[0] == 3:
        mrope_section = rope_scaling.get("mrope_section",
                                         None) if rope_scaling else None
        # NOTE: We assume mrope_section is always available
        # as Qwen2.5-VL is the only model using mrope
        assert mrope_section is not None

        split_indices = [mrope_section[0], mrope_section[0] + mrope_section[1]]

        # Indices for the features to be rotated (first half of head_dim)
        all_freq_indices = jnp.arange(head_dim // 2)

        # Split the indices according to mrope_section. This is valid because split_indices are static.
        freq_indices_split = jnp.split(all_freq_indices, split_indices)
        # freq_indices_split is a list of 3 JAX arrays.

        cos_list = []
        sin_list = []

        for i in range(3):  # For each of the 3 position dimensions
            current_indices = freq_indices_split[i]

            if current_indices.size == 0:
                # This section is empty, skip.
                continue

            # inv_freq shape: (mrope_section[i],)
            inv_freq = 1.0 / (rope_theta**(
                current_indices.astype(jnp.float32) * 2.0 / head_dim))

            # positions[i]: (seq_len,)
            # freqs shape: (seq_len, mrope_section[i])
            freqs = jnp.outer(positions[i].astype(jnp.float32), inv_freq)

            cos_list.append(jnp.cos(freqs))
            sin_list.append(jnp.sin(freqs))

        # Concatenate along the feature dimension
        # cos, sin shape: (seq_len, head_dim//2)
        cos = jnp.concatenate(cos_list, axis=1)
        sin = jnp.concatenate(sin_list, axis=1)

        # Add num_heads dimension for broadcasting
        cos = cos[:, jnp.newaxis, :]  # Shape: (seq_len, 1, head_dim//2)
        sin = sin[:, jnp.newaxis, :]  # Shape: (seq_len, 1, head_dim//2)

        # Apply rotation
        inputs_real = inputs[..., :head_dim // 2]
        inputs_imag = inputs[..., head_dim // 2:head_dim]

        outputs_real = inputs_real * cos - inputs_imag * sin
        outputs_imag = inputs_real * sin + inputs_imag * cos

        out = jnp.concatenate([outputs_real, outputs_imag], axis=-1)

    # The positions tensor is multidimensional for vision (hence >= 2) and
    # the value of the last axis should be 2 as the tensor holds the pre-computed
    # rotation factors (cosine and sine values) for each position.
    elif positions.ndim >= 2 and positions.shape[-1] == 2:  #VISION RoPE
        # positions = freqs_cis_stacked: (S, D_rot, 2)
        # Unstack to get the complex rotation factors (cos + i*sin) in a trace-compatible way.
        cos = jax.lax.slice_in_dim(positions, 0, 1, axis=-1)
        sin = jax.lax.slice_in_dim(positions, 1, 2, axis=-1)

        # Squeeze the trailing 1 dimension to get (S, D_rot)
        cos = jnp.squeeze(cos, axis=-1)
        sin = jnp.squeeze(sin, axis=-1)

        seq_len_input = inputs.shape[0]
        seq_len_pos = cos.shape[0]

        if seq_len_input != seq_len_pos:
            # Calculate how many images/blocks are in the input
            num_repeats = seq_len_input // seq_len_pos
            cos = jnp.tile(cos, (num_repeats, 1))
            sin = jnp.tile(sin, (num_repeats, 1))

        # Reshape to (S, 1, D_rot) for broadcasting over heads
        cos = cos[:, jnp.newaxis, :]
        sin = sin[:, jnp.newaxis, :]

        cos_f32 = cos.astype(jnp.float32)
        sin_f32 = sin.astype(jnp.float32)
        inputs_f32 = inputs.astype(jnp.float32)

        if rope_input_ordering == "interleaved":
            shape_pre = inputs_f32.shape[:-1]
            inputs_reshaped = inputs_f32.reshape(*shape_pre, -1, 2)

            inputs_real = inputs_reshaped[..., 0]
            inputs_imag = inputs_reshaped[..., 1]

            outputs_real = inputs_real * cos_f32 - inputs_imag * sin_f32
            outputs_imag = inputs_real * sin_f32 + inputs_imag * cos_f32

            out_stacked = jnp.stack([outputs_real, outputs_imag], axis=-1)
            out = out_stacked.reshape(inputs.shape)

        else:
            inputs_real = inputs_f32[..., :head_dim // 2]
            inputs_imag = inputs_f32[..., head_dim // 2:head_dim]

            outputs_real = inputs_real * cos_f32 - inputs_imag * sin_f32
            outputs_imag = inputs_real * sin_f32 + inputs_imag * cos_f32

            out = jnp.concatenate([outputs_real, outputs_imag], axis=-1)

    # Standard RoPE or Partial RoPE
    else:
        # Calculate inverse frequencies (timescale)
        rope_angles = int(rope_proportion * head_dim // 2)
        nope_angles = head_dim // 2 - rope_angles

        fraction = 2 * jnp.arange(0, rope_angles, dtype=jnp.float32) / head_dim
        timescale = 1.0 / (rope_theta**fraction)

        # Apply scaling if provided
        if rope_scaling:
            timescale = apply_rope_scaling(timescale, rope_scaling)

        if nope_angles > 0:
            timescale = jnp.pad(timescale, (0, nope_angles),
                                mode="constant",
                                constant_values=0.0)

        # Prepare for rotation by calculating sin and cos values
        # `sinusoid_inp` gets shape (batch * seq_len, head_dim/2)
        sinusoid_inp = positions[..., jnp.newaxis].astype(
            jnp.float32) * timescale[jnp.newaxis, :]

        # Broadcast over the 'heads' dimension, assuming shape (batch*seq, heads, head_dim)
        sinusoid_inp = sinusoid_inp[:, jnp.newaxis, ...]
        sin = jnp.sin(sinusoid_inp)
        cos = jnp.cos(sinusoid_inp)

        if rope_input_ordering == "interleaved":
            # Reshape to group adjacent features for rotation, matching new_apply_rope
            rotary_inputs = inputs[
                ..., :head_dim]  # Take just the non-padded amount.
            reshaped_inputs = rotary_inputs.reshape(*rotary_inputs.shape[:-1],
                                                    -1, 2)

            # Apply the rotation
            first_half = reshaped_inputs[..., 0]
            second_half = reshaped_inputs[..., 1]
        else:
            first_half = inputs[..., :head_dim // 2]
            second_half = inputs[..., head_dim // 2:head_dim]

        first_part = first_half * cos - second_half * sin
        second_part = second_half * cos + first_half * sin

        # Combine the rotated parts and reshape back
        if rope_input_ordering == "interleaved":
            out_stacked = jnp.stack([first_part, second_part], axis=-1)
            out = out_stacked.reshape(rotary_inputs.shape)
        else:
            out = jnp.concatenate([first_part, second_part], axis=-1)

    # If the original input was padded, pad the output with zeros to match.
    padded_head_dim = inputs.shape[-1]
    if padded_head_dim > head_dim:
        pad_width = padded_head_dim - head_dim
        pad_config = [(0, 0)] * (out.ndim - 1) + [(0, pad_width)]
        out = jnp.pad(out, pad_config)

    return out.astype(inputs.dtype)


def apply_longrope(
    inputs: jax.Array,
    positions: jax.Array,
    head_dim: int,
    rope_scaling: Dict[str, Any],
    original_max_position_embeddings: int,
    max_position_embeddings: int,
    rope_theta: float = 10000,
) -> jax.Array:
    # LongRoPE implementation specific to Phi-3
    # Implementation based on https://github.com/huggingface/transformers/blob/main/src/transformers/models/phi3/modeling_phi3.py#L197-L235

    scale = max_position_embeddings / original_max_position_embeddings
    if scale <= 1.0:
        mscale = 1.0
    else:
        mscale = jnp.sqrt(1 + (jnp.log(scale) /
                               jnp.log(original_max_position_embeddings)))

    seq_len = inputs.shape[0]
    if seq_len > original_max_position_embeddings:
        long_factor = jnp.array(rope_scaling.get("long_factor"))
        timescale = 1.0 / (long_factor * (rope_theta**(
            (2 * jnp.arange(0, head_dim // 2, dtype=jnp.float32)) / head_dim)))
    else:
        short_factor = jnp.array(rope_scaling.get("short_factor"))
        timescale = 1.0 / (short_factor * (rope_theta**(
            (2 * jnp.arange(0, head_dim // 2, dtype=jnp.float32)) / head_dim)))

    # Calculate RoPE positions
    sinusoid_inp = positions[..., jnp.newaxis].astype(
        jnp.float32) * timescale[jnp.newaxis, :]
    sinusoid_inp = sinusoid_inp[:, jnp.newaxis, ...]
    sin = jnp.sin(sinusoid_inp) * mscale
    cos = jnp.cos(sinusoid_inp) * mscale

    # Padding logic
    padded_head_dim = inputs.shape[-1]

    # Apply RoPE mechanism
    first_half = inputs[..., :head_dim // 2]
    second_half = inputs[..., head_dim // 2:head_dim]

    first_part = first_half * cos - second_half * sin
    second_part = second_half * cos + first_half * sin
    out = jnp.concatenate([first_part, second_part], axis=-1)

    if padded_head_dim > head_dim:
        out = jnp.pad(out, ((0, 0), (0, 0), (0, padded_head_dim - head_dim)))

    return out.astype(inputs.dtype)


def apply_rope_scaling(freqs: jax.Array, rope_scaling: Dict[str,
                                                            Any]) -> jax.Array:
    # Values obtained from grid search
    scale_factor = rope_scaling.get("scale_factor", 8.0)
    low_freq_factor = rope_scaling.get("low_freq_factor", 1.0)
    high_freq_factor = rope_scaling.get("high_freq_factor", 4.0)
    old_context_len = rope_scaling.get("original_max_position_embeddings",
                                       8192)

    low_freq_wavelen = old_context_len / low_freq_factor
    high_freq_wavelen = old_context_len / high_freq_factor

    wavelen = 2 * math.pi / freqs
    smooth = (old_context_len / wavelen -
              low_freq_factor) / (high_freq_factor - low_freq_factor)

    high_freqs = jnp.where(wavelen < high_freq_wavelen, freqs, 0)
    low_freqs = jnp.where(wavelen > low_freq_wavelen, freqs / scale_factor, 0)
    mid_freqs = jnp.where(
        (wavelen >= high_freq_wavelen) & (wavelen <= low_freq_wavelen),
        (1 - smooth) * freqs / scale_factor + smooth * freqs,
        0,
    )
    new_freqs = high_freqs + low_freqs + mid_freqs
    return new_freqs



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Bridge the torch gdn_attention_core op for gated deltanet attention TPU impl

"""
import functools
from typing import Optional, Tuple

import jax
import jax.numpy as jnp
import torch
from jax.sharding import PartitionSpec as P
from torchax.interop import jax_view, torch_view
from vllm.forward_context import get_forward_context

from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.layers.common.utils import \
    reorder_concatenated_tensor_for_sharding
from tpu_inference.layers.vllm.ops.ragged_conv1d_jax import ragged_conv1d
from tpu_inference.layers.vllm.ops.ragged_gated_delta_rule_jax import \
    ragged_gated_delta_rule
from tpu_inference.models.vllm.vllm_model_wrapper_context import \
    get_vllm_model_wrapper_context


def run_jax_gdn_attention_local(
    mixed_qkv: jnp.ndarray,
    b: jnp.ndarray,
    a: jnp.ndarray,
    conv_state: jnp.ndarray,
    recurrent_state: jnp.ndarray,
    conv_weight: jnp.ndarray,
    conv_bias: Optional[jnp.ndarray],
    A_log: jnp.ndarray,
    dt_bias: jnp.ndarray,
    query_start_loc: jnp.ndarray,
    state_indices: jnp.ndarray,
    n_kq: int,
    n_v: int,
    d_k: int,
    d_v: int,
    kernel_size: int,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
    """Runs the local JAX GDN attention mechanism with combined QKV tensors.

    Args:
        mixed_qkv: Combined QKV tensor of shape `(num_tokens, dim)`.
        b: B tensor of shape `(num_tokens, n_v)`.
        a: A tensor of shape `(num_tokens, n_v)`.
        conv_state: Combined convolutional state of shape `(max_reqs, kernel_size
          - 1, dim)`.
        recurrent_state: Recurrent state of shape `(max_reqs, n_v, d_k, d_v)`.
        conv_weight: Combined convolutional weight of shape `(dim, 1,
          kernel_size)`.
        conv_bias: Optional combined convolutional bias of shape `(dim,)`.
        A_log: Log of A parameter of shape `(n_v,)`.
        dt_bias: Delta T bias of shape `(n_v,)`.
        query_start_loc: Tensor of shape `(num_seqs + 1,)` with start locations of
          each sequence.
        state_indices: Tensor of shape `(max_reqs,)` mapping request index to
          state index.
        n_kq: Number of key/query heads.
        n_v: Number of value heads.
        d_k: Dimension of key.
        d_v: Dimension of value.
        kernel_size: Convolution kernel size.

    Returns:
        A tuple containing the new states and the output.
        - A tuple of (new_conv_state, new_recurrent_state).
        - The output tensor of shape `(num_tokens, n_v * d_v)`.
    """
    # Ensure query_start_loc is monotonically increasing to handle padded slots
    query_start_loc = jnp.maximum.accumulate(query_start_loc)

    out_mixed_qkv, new_conv_state = ragged_conv1d(
        mixed_qkv,
        conv_state,
        conv_weight,
        conv_bias,
        query_start_loc,
        state_indices,
        kernel_size,
    )

    out_mixed_qkv = jax.nn.silu(out_mixed_qkv)

    new_recurrent_state, output = ragged_gated_delta_rule(
        out_mixed_qkv,
        b,
        a,
        recurrent_state,
        A_log,
        dt_bias,
        query_start_loc,
        state_indices,
        n_kq,
        n_v,
        d_k,
        d_v,
    )

    return (new_conv_state, new_recurrent_state), output


def run_jax_gdn_attention(
    j_mixed_qkv: jnp.ndarray,
    j_b: jnp.ndarray,
    j_a: jnp.ndarray,
    conv_state: jnp.ndarray,
    recurrent_state: jnp.ndarray,
    j_conv_weight: jnp.ndarray,
    j_conv_bias: Optional[jnp.ndarray],
    j_A_log: jnp.ndarray,
    j_dt_bias: jnp.ndarray,
    state_indices: jnp.ndarray,
    query_start_loc: jnp.ndarray,
    n_kq: int,
    n_v: int,
    d_k: int,
    d_v: int,
    kernel_size: int,
    mesh: jax.sharding.Mesh,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
    """Runs the Jax GDN attention mechanism.

    Args:
        j_mixed_qkv: Input tensor of shape `(num_tokens, dim)`.
        j_b: Input tensor of shape `(num_tokens, n_v)`.
        j_a: Input tensor of shape `(num_tokens, n_v)`.
        conv_state: Convolutional state tensor of shape `(max_reqs, kernel_size -
          1, dim)`.
        recurrent_state: Recurrent state tensor of shape `(max_reqs, n_v, d_k,
          d_v)`.
        j_conv_weight: Convolutional weight tensor of shape `(dim, 1,
          kernel_size)`.
        j_conv_bias: Optional convolutional bias tensor of shape `(dim,)`.
        j_A_log: Log of A parameter tensor of shape `(n_v,)`.
        j_dt_bias: Delta T bias tensor of shape `(n_v,)`.
        state_indices: Tensor of shape `(max_reqs,)` mapping request index to
          state index.
        query_start_loc: Tensor of shape `(num_seqs + 1,)` with start locations of
          each sequence.
        n_kq: Number of key/query heads.
        n_v: Number of value heads.
        d_k: Dimension of key.
        d_v: Dimension of value.
        kernel_size: Convolution kernel size.
        mesh: The device mesh for distributed computation.

    Returns:
        A tuple containing the new states and the output.
        - A tuple of (new_conv_state, new_recurrent_state).
          - new_conv_state: `(max_reqs, kernel_size - 1, dim)`
          - new_recurrent_state: `(max_reqs, n_v, d_k, d_v)`
        - The output tensor of shape `(num_tokens, n_v * d_v)`.
    """
    in_specs = (
        P(None, ShardingAxisName.ATTN_HEAD),  # j_mixed_qkv
        P(None, ShardingAxisName.ATTN_HEAD),  # j_b
        P(None, ShardingAxisName.ATTN_HEAD),  # j_a
        P(None, None, ShardingAxisName.ATTN_HEAD),  # conv_state
        P(None, ShardingAxisName.ATTN_HEAD, None, None),  # recurrent_state
        P(ShardingAxisName.ATTN_HEAD, None, None),  # j_conv_weight
        P(ShardingAxisName.ATTN_HEAD)
        if j_conv_bias is not None else None,  # j_conv_bias
        P(ShardingAxisName.ATTN_HEAD),  # j_A_log
        P(ShardingAxisName.ATTN_HEAD),  # j_dt_bias
        P(),  # query_start_loc
        P(),  # state_indices
    )

    out_specs = (
        (
            P(None, None, ShardingAxisName.ATTN_HEAD),  # new_conv_state
            P(None, ShardingAxisName.ATTN_HEAD, None,
              None),  # new_recurrent_state
        ),
        P(None, ShardingAxisName.ATTN_HEAD),  # output
    )

    tp_size = mesh.shape[ShardingAxisName.ATTN_HEAD]

    p_run_jax_gdn_attention_local = functools.partial(
        run_jax_gdn_attention_local,
        n_kq=n_kq // tp_size,
        n_v=n_v // tp_size,
        d_k=d_k,
        d_v=d_v,
        kernel_size=kernel_size,
    )

    mapped_fn = jax.shard_map(
        p_run_jax_gdn_attention_local,
        mesh=mesh,
        in_specs=in_specs,
        out_specs=out_specs,
    )

    (new_conv_state, new_recurrent_state), output = mapped_fn(
        j_mixed_qkv,
        j_b,
        j_a,
        conv_state,
        recurrent_state,
        j_conv_weight,
        j_conv_bias,
        j_A_log,
        j_dt_bias,
        query_start_loc,
        state_indices,
    )

    return (new_conv_state, new_recurrent_state), output


def gdn_attention_core_tpu(
    mixed_qkv: torch.Tensor,
    b: torch.Tensor,
    a: torch.Tensor,
    core_attn_out: torch.Tensor,
    layer_name: str,
    mesh: jax.sharding.Mesh,
) -> None:
    """
    This acts as main bridge between PyTorch and JAX for the GDN core attention.
    Uses a robust, token-by-token scan to inherently handle any mix of
    ragged prefill and decode sequences without dynamic shape compilation errors.

    Some key details:
    1. Cache Mapping: We'll read vLLM's  `block_tables` and `query_start_loc`
       and translate them into static index arrays (`req_indices` and `state_indices`).
    2. JAX Scan: We use `jax.lax.scan` to perform a robust, token-by-token loop over
       the flat inputs. This allows us to handle ANY mix of prefill and decode tokens
       in a single compiled XLA graph.
    3. Conditional Updates: The `valid_mask` ensures that padded dummy tokens
       (used to keep the tensor shape static) do not corrupt the recurrent state
       in the cache.
    """
    fc = get_forward_context()
    attn_metadata = fc.attn_metadata[layer_name]

    layer_module = fc.no_compile_layers[layer_name]
    vllm_context = get_vllm_model_wrapper_context()

    n_kq = layer_module.num_k_heads
    n_v = layer_module.num_v_heads
    d_k = layer_module.head_k_dim
    d_v = layer_module.head_v_dim
    kernel_size = layer_module.conv_kernel_size

    j_mixed_qkv = jax_view(mixed_qkv)  # [num_tokens, dim]
    j_b = jax_view(b)
    j_a = jax_view(a)

    j_conv_weight = jax_view(layer_module.conv1d.weight)
    j_conv_bias = jax_view(layer_module.conv1d.bias
                           ) if layer_module.conv1d.bias is not None else None
    j_A_log = jax_view(layer_module.A_log)
    j_dt_bias = jax_view(layer_module.dt_bias)

    # The j_mixed_qkv and j_conv_weight are not in an interleaved layout.
    # E.g. they are in [Q Q | K K | V V] layout. We need [Q K | Q K | Q K] layout.
    # Use reorder_concatenated_tensor_for_sharding to reorder into correct layout
    key_dim = n_kq * d_k
    value_dim = n_v * d_v
    tp_size = mesh.shape[ShardingAxisName.ATTN_HEAD]
    j_mixed_qkv = reorder_concatenated_tensor_for_sharding(
        j_mixed_qkv, [key_dim, key_dim, value_dim], tp_size, -1)
    j_conv_weight = reorder_concatenated_tensor_for_sharding(
        j_conv_weight, [key_dim, key_dim, value_dim], tp_size, 0)

    layer_idx = vllm_context.layer_name_to_kvcache_index[layer_name]
    conv_state, recurrent_state = vllm_context.kv_caches[layer_idx]

    # Map physical cache blocks
    flat_block_tables = jax_view(attn_metadata.block_tables)
    max_reqs = attn_metadata.seq_lens.shape[0]
    max_blocks_per_req = flat_block_tables.shape[0] // max_reqs
    block_tables_2d = jnp.reshape(flat_block_tables,
                                  (max_reqs, max_blocks_per_req))
    state_indices = block_tables_2d[:, 0].astype(jnp.int32)

    # Map tokens to their respective requests
    q_loc = jax_view(attn_metadata.query_start_loc)

    (new_conv_state,
     new_recurrent_state), j_output = run_jax_gdn_attention(j_mixed_qkv,
                                                            j_b,
                                                            j_a,
                                                            conv_state,
                                                            recurrent_state,
                                                            j_conv_weight,
                                                            j_conv_bias,
                                                            j_A_log,
                                                            j_dt_bias,
                                                            state_indices,
                                                            q_loc,
                                                            n_kq,
                                                            n_v,
                                                            d_k,
                                                            d_v,
                                                            kernel_size,
                                                            mesh=mesh)

    vllm_context.kv_caches[layer_idx] = (new_conv_state, new_recurrent_state)

    j_output_flat = j_output.reshape(core_attn_out.shape)
    core_attn_out.copy_(torch_view(j_output_flat))


def gdn_in_proj_tpu(
    hidden_states: torch.Tensor,
    qkvz_size: int,
    ba_size: int,
    prefix: str,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Torch OP replacement for gdn_in_proj.
    Executes the underlying linear layers (in_proj_qkvz and in_proj_ba) directly
    to mimic the old/non-kernel execution path that was reverted in:
    https://github.com/vllm-project/vllm/pull/36795

    Args:
        hidden_states: Tensor of shape (num_tokens, hidden_size)
        qkvz_size: int, unused but needed for signature
        ba_size: int, unused but needed for signature
        prefix: str
    Returns:
        mixed_qkvz: Tensor of shape (num_tokens, qkvz_size)
        ba: Tensor of shape (num_tokens, ba_size)
    """
    fc = get_forward_context()
    # The 'prefix' argument perfectly matches the key used to register the module
    layer_module = fc.no_compile_layers[prefix]

    # Run the original projections instead of the fused C++ kernel
    mixed_qkvz, _ = layer_module.in_proj_qkvz(hidden_states)
    ba, _ = layer_module.in_proj_ba(hidden_states)

    return mixed_qkvz, ba


def apply_gated_delta_net_torch_ops_patch(mesh: jax.sharding.Mesh) -> None:
    """
    This is a patch to inject the `gdn_attention_core` op so the
    Torch/GPU  kernel is bypassed in favor of the TPU kernel
    here:
    https://github.com/vllm-project/vllm/blob/697e4ff3528c72806a4d00ed9b7581332b9efd43/vllm/model_executor/models/qwen3_next.py#L671

    """
    try:
        import vllm.model_executor.models.qwen3_next  # noqa: F401
    except ImportError:
        pass

    # Ensure the op exists in the namespace, which initializes the OpOverloadPacket
    if hasattr(torch.ops, "vllm") and hasattr(torch.ops.vllm,
                                              "gdn_attention_core"):
        # dummy call to ensure the op is registered
        torch.ops.vllm.gdn_attention_core = functools.partial(
            gdn_attention_core_tpu, mesh=mesh)

    if hasattr(torch.ops.vllm, "gdn_in_proj"):
        # dummy call to ensure the op is registered
        torch.ops.vllm.gdn_in_proj = gdn_in_proj_tpu



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ragged convolution Jax implementation."""

import jax.numpy as jnp


def ragged_conv1d(
    x,
    conv_state,
    conv_weight,
    conv_bias,
    query_start_loc,
    state_indices,
    kernel_size,
):
    """Applies 1D convolution over ragged sequences and updates state.

    Args:
      x: Input tensor of shape `(num_tokens, dim)`.
      conv_state: Combined convolutional state of shape `(max_reqs, kernel_size -
        1, dim)`.
      conv_weight: Convolutional weight of shape `(dim, 1, kernel_size)`.
      conv_bias: Optional convolutional bias of shape `(dim,)`.
      query_start_loc: Tensor of shape `(num_seqs + 1,)` containing the start
        indices of each sequence, with the last element being the total number of
        valid tokens.
      state_indices: Tensor of shape `(max_reqs,)` mapping request index to state
        index.
      kernel_size: The size of the convolution kernel.
  
    Returns:
      A tuple containing:
      - output: The output tensor of shape `(num_tokens, dim)`.
      - updated_conv_state: The updated convolutional state of shape `(max_reqs,
        kernel_size - 1, dim)`.
    """
    num_tokens = x.shape[0]
    max_reqs = state_indices.shape[0]
    token_idx = jnp.arange(num_tokens)

    req_indices = (
        jnp.sum(token_idx[:, None] >= query_start_loc[None, :], axis=1) - 1)
    req_indices = jnp.clip(req_indices, 0, max_reqs - 1)
    local_indices = token_idx - query_start_loc[req_indices]

    lengths = query_start_loc[1:] - query_start_loc[:-1]

    # 1. Compute Convolution
    out = jnp.zeros_like(x)
    w = conv_weight[:, 0, :].T  # (K, C)

    gathered_state = conv_state[state_indices]  # (max_reqs, K-1, C)

    for k in range(kernel_size):
        mask = local_indices >= k

        idx_x = jnp.clip(token_idx - k, 0, num_tokens - 1)
        idx_state_t = (kernel_size - 1) + (local_indices - k)

        x_tokens = x[idx_x]
        state_tokens = gathered_state[req_indices, idx_state_t]

        token_k = jnp.where(mask[:, None], x_tokens, state_tokens)
        out = out + token_k * w[kernel_size - 1 - k]

    if conv_bias is not None:
        out = out + conv_bias[jnp.newaxis, :]

    # 2. Update State
    padded_lengths = jnp.zeros(max_reqs, dtype=jnp.int32)
    padded_lengths = padded_lengths.at[:lengths.shape[0]].set(lengths)

    padded_q_loc = jnp.zeros(max_reqs + 1, dtype=jnp.int32)
    padded_q_loc = padded_q_loc.at[:query_start_loc.shape[0]].set(
        query_start_loc)
    padded_q_loc = padded_q_loc.at[query_start_loc.shape[0]:].set(
        query_start_loc[-1])

    r_grid = jnp.arange(max_reqs)[:, None]
    j_grid = jnp.arange(kernel_size - 1)[None, :]

    v_i = padded_lengths[:, None] + j_grid
    is_from_old_state = v_i < kernel_size - 1

    idx_state = jnp.where(is_from_old_state, v_i, 0)
    idx_x = padded_q_loc[r_grid + 1] + j_grid - (kernel_size - 1)
    idx_x = jnp.clip(idx_x, 0, num_tokens - 1)

    new_state_extracted = jnp.where(is_from_old_state[..., None],
                                    gathered_state[r_grid,
                                                   idx_state], x[idx_x])

    valid_seq_mask = jnp.arange(max_reqs) < lengths.shape[0]
    updated_conv_state = conv_state.at[state_indices].set(
        jnp.where(
            valid_seq_mask[:, None, None],
            new_state_extracted,
            conv_state[state_indices],
        ))

    return out, updated_conv_state



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ragged gated delta rule JAX implementation."""

from typing import Optional, Tuple

import jax
import jax.numpy as jnp


def _l2_normalize(x: jnp.ndarray, eps: float = 1e-6) -> jnp.ndarray:
    """L2 normalize along last dimension.

    Args:
        x: input to normalize
        eps: epsilon for numerical stability

    Returns:
        normalized x
    """
    norm = jnp.sqrt(jnp.sum(x * x, axis=-1, keepdims=True) + eps)
    return x / norm


def _recurrent_gated_delta_rule_step(
    query: jnp.ndarray,
    key: jnp.ndarray,
    value: jnp.ndarray,
    g: jnp.ndarray,
    beta: jnp.ndarray,
    state: Optional[jnp.ndarray] = None,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Single-step recurrent update for decode.

    Args:
      query: Tensor of shape `(B, H, T, d_k)`. Current implementation assumes
        `T=1`.
      key: Tensor of shape `(B, H, T, d_k)`. Current implementation assumes `T=1`.
      value: Tensor of shape `(B, H, T, d_v)`. Current implementation assumes
        `T=1`.
      g: Tensor of shape `(B, H, T)`. Current implementation assumes `T=1`.
      beta: Tensor of shape `(B, H, T)`. Current implementation assumes `T=1`.
      state: Optional initial recurrent state of shape `(B, H, d_k, d_v)`.

    Returns:
      A tuple containing:
      - output: The output tensor of shape `(B, H, T, d_v)`.
      - new_state: The updated recurrent state of shape `(B, H, d_k, d_v)`.
    """
    B, H, T, d_k = query.shape
    d_v = value.shape[-1]

    if state is None:
        state = jnp.zeros((B, H, d_k, d_v), dtype=query.dtype)

    q = query[:, :, 0]
    k = key[:, :, 0]
    v = value[:, :, 0]
    beta_val = beta[:, :, 0]
    g_val = g[:, :, 0]

    scale = d_k**-0.5
    q = q * scale

    # v_diff = v - e^g * (k @ state)
    k_state = jnp.einsum("bhd, bhdm -> bhm", k, state)
    v_diff = v - jnp.exp(g_val)[..., None] * k_state

    # v_new = beta * v_diff
    v_new = beta_val[..., None] * v_diff

    # out = e^g * (q @ state) + (q . k) * v_new
    q_state = jnp.einsum("bhd, bhdm -> bhm", q, state)
    q_k = jnp.sum(q * k, axis=-1, keepdims=True)

    out = jnp.exp(g_val)[..., None] * q_state + q_k * v_new

    # s_new = state * exp(g) + k outer v_new
    k_v_new = jnp.einsum("bhd, bhm -> bhdm", k, v_new)
    new_state = state * jnp.exp(g_val)[..., None, None] + k_v_new

    return out[:, :, None, :], new_state


def ragged_gated_delta_rule(
    mixed_qkv,
    b,
    a,
    recurrent_state,
    A_log,
    dt_bias,
    query_start_loc,
    state_indices,
    n_kq,
    n_v,
    d_k,
    d_v,
):
    """Applies the gated delta rule over ragged sequences and updates recurrent state.

    Args:
      mixed_qkv: Combined QKV tensor of shape `(num_tokens, 2 * n_kq * d_k + n_v *
        d_v)`.
      b: B tensor of shape `(num_tokens, n_v)`.
      a: A tensor of shape `(num_tokens, n_v)`.
      recurrent_state: Recurrent state of shape `(max_reqs, n_v, d_k, d_v)`.
      A_log: Log of A parameter of shape `(n_v,)`.
      dt_bias: Delta T bias of shape `(n_v,)`.
      query_start_loc: Tensor of shape `(num_seqs + 1,)` containing the start
        indices of each sequence, with the last element being the total number of
        valid tokens.
      state_indices: Tensor of shape `(max_reqs,)` mapping request index to state
        index.
      n_kq: Number of key/query heads.
      n_v: Number of value heads.
      d_k: Dimension of key.
      d_v: Dimension of value.

    Returns:
      A tuple containing:
      - updated_recurrent_state: The updated recurrent state of shape `(max_reqs,
        n_v, d_k, d_v)`.
      - output: The output tensor of shape `(num_tokens, n_v * d_v)`.
    """
    num_tokens = mixed_qkv.shape[0]
    key_dim = n_kq * d_k
    query = mixed_qkv[..., :key_dim]
    key = mixed_qkv[..., key_dim:key_dim * 2]
    value = mixed_qkv[..., key_dim * 2:]
    max_reqs = state_indices.shape[0]
    token_idx = jnp.arange(num_tokens)

    req_indices = (
        jnp.sum(token_idx[:, None] >= query_start_loc[None, :], axis=1) - 1)
    req_indices = jnp.clip(req_indices, 0, max_reqs - 1)
    valid_mask = token_idx < query_start_loc[-1]

    def scan_fn(carry, xs):
        recurrent_state_all = carry
        (
            curr_q,
            curr_k,
            curr_v,
            curr_b,
            curr_a,
            request_index,
            is_valid_token,
        ) = xs

        curr_q = curr_q[None, None, :]
        curr_k = curr_k[None, None, :]
        curr_v = curr_v[None, None, :]
        curr_b = curr_b[None, None, :]
        curr_a = curr_a[None, None, :]

        state_index = state_indices[request_index]
        recurrent_state = recurrent_state_all[state_index][None, ...]

        B, T = 1, 1
        query_reshaped = curr_q.reshape(B, T, n_kq, d_k)
        key_reshaped = curr_k.reshape(B, T, n_kq, d_k)
        value_reshaped = curr_v.reshape(B, T, n_v, d_v)

        beta = jax.nn.sigmoid(curr_b)
        g = -jnp.exp(A_log.astype(jnp.float32)) * jax.nn.softplus(
            curr_a.astype(jnp.float32) + dt_bias.astype(jnp.float32))

        repeat_factor = n_v // n_kq
        if repeat_factor > 1:
            query_reshaped = jnp.repeat(query_reshaped, repeat_factor, axis=2)
            key_reshaped = jnp.repeat(key_reshaped, repeat_factor, axis=2)

        query_reshaped = jnp.transpose(query_reshaped,
                                       (0, 2, 1, 3)).astype(jnp.float32)
        key_reshaped = jnp.transpose(key_reshaped,
                                     (0, 2, 1, 3)).astype(jnp.float32)
        value_reshaped = jnp.transpose(value_reshaped,
                                       (0, 2, 1, 3)).astype(jnp.float32)
        beta = jnp.transpose(beta, (0, 2, 1)).astype(jnp.float32)
        g = jnp.transpose(g, (0, 2, 1)).astype(jnp.float32)

        query_reshaped = _l2_normalize(query_reshaped)
        key_reshaped = _l2_normalize(key_reshaped)

        output, new_recurrent_state = _recurrent_gated_delta_rule_step(
            query_reshaped,
            key_reshaped,
            value_reshaped,
            g,
            beta,
            state=recurrent_state,
        )

        output = jnp.transpose(output, (0, 2, 1, 3)).astype(query.dtype)
        output = output.reshape(B, T, -1)

        recurrent_state_all = jnp.where(
            is_valid_token,
            recurrent_state_all.at[state_index].set(new_recurrent_state[0]),
            recurrent_state_all,
        )

        return recurrent_state_all, output[0, 0]

    carry_init = recurrent_state
    xs = (query, key, value, b, a, req_indices, valid_mask)

    new_recurrent_state, output = jax.lax.scan(scan_fn, carry_init, xs)
    return new_recurrent_state, output
